# fritzconnection library:
#
# The modules in the lib folder are helper modules to access some services
# of the FritzBox. They can also be used as examples on how to write
# applications using fritzconnection.
